%% Should be the first file when catenating the
%% files comprising the source_debug module

%% make sure this matches load.pl
:- module(source_debug,
          [konsult/1,
           %% profile entrypoints
           profile_enable/2,
           reset_profile_data/0, reset_profile_data/1,
           dump_profile_data/0, dump_profile_data/1
          ]).

copyright('Copyright (c) 1995,1996 Per Mildner, all rights reserved').
%%% 960901 Do not change this file!
%%% A newer version is in ~/prolog/rtg/generic_consult.pl

%%% Copyright (c) 1995 Per Mildner, all rights reserved.
%%% email: Per.Mildner@csd.uu.se
%%% A Source debug hack. Hopefully the last prolog program I develop
%%% without access to a source debugger :-)
%%%
%%% To use:
%%% konsult/1 same arguments as prolog:consult/1
%%% Together with the latest version of the sicstus-debug-mode package
%%% this will follow the source code while using the SICStus debugger.
%%%
%%% Currently rehacking to provide a more generic process-prolog-file engine
%%%
%%% Does not do modules right yet
%%% Does not support load_context
%%% Does not support any of the info about which predicate goes in
%%% which file etc.
%%% Does not do the right thing with multifile declarations
%%%
%%% ToDo: (also look for !!! in the code)
%%% . handle modules right. Currently ignores public predicate list
%%% . do goal_expansion
%%% . meta_predicate, multifile etc. see handle_declaration
%%%
%%% . DONE: make it its own module
%%% . DONE: change from user:pos/2 as marker to something such as
%%%   src_dbg:'$P'/2 (it should look different)
%%% . DONE: Allow breakpoints at a File/Line, probably doable by conditional
%%%   break on pos/2
%%%


:- ensure_loaded(library(system)).

:- meta_predicate konsult(:).


:- dynamic konsult_context/2.

konsult_context(dummy, _Value) :- fail.          % dummy



konsult(X) :-
   prolog_flag(typein_module,M),
   act_init(Act),
   on_exception(Excp,
     konsult(X, M, Act),
     (
         %% retractall(konsult_context(_,_)),
         prolog_flag(typein_module,_,M),
         raise_exception(Excp)
     )
   ),
   prolog_flag(typein_module,_,M).


konsult(X, _M, _Act) :- var(X), !, fail.              % !!! should give error

konsult(M:X, _M, Act) :- !,
   konsult(X, M, Act).

konsult([F|Fs], M, Act) :- !,
   konsult_files([F|Fs], M, Act).

konsult(F, M, Act0) :-
   konsult_file(F, M, Act0, Act),
   act_final(Act).


konsult_files([], _M, Act) :-
   act_final(Act).

konsult_files([F|Fs], M, Act0) :-
   konsult_file(F, M, Act0, Act1),
   konsult_files(Fs, M, Act1).


konsult_file(X, _M, _Act, _) :- var(X), !, fail.          % !!! should give error

konsult_file(-File, M, Act0, Act) :- !,
   konsult_file(File, M, Act0, Act).

konsult_file(M:File, _M, Act0, Act) :- !,
   konsult_file(File, M, Act0, Act).


konsult_file(File0, M, Act0, Act) :-
   statistics(runtime, [T0|_]),
   prolog_flag(typein_module,M0),
   module(M),
   working_directory(WD,WD),
   absolute_file_name(File0,File),              % will suffix .pl if needed
   format('~N{konsulting ~w}~n', [File]),

   file_directory(File, Dir),

   (
       on_exception(Excp,
          see(File),
          (
              ( Excp = existence_error(see(_),1,file,Culp,Msg) ->
                  print_message(error, existence_error(konsult(_),1,file,Culp,Msg))
              ; print_message(error, Excp)
              ),
              raise_exception('konsult aborted')
          )),
        %% Do this when we know that see/1 succeded.
        %% otherwise we will get the nonexisting file as the context
        %% for any nonexisting-file error
        asserta(konsult_context(file, File), R1),
        asserta(konsult_context(directory, Dir), R2),
        asserta(konsult_context(module, M), R3),
        %% no stream or term_position, who needs it?


       on_exception(Excp,
          read_klauses_stream(File, Dir, FileSegments, M, Act0, Act1),
          (
              seen,
              %% print message before removeing load context!
              ( Excp = syntax_error(_Goal,Between,Msg,Tokens,AfterError) ->
                  %% remove error context
                  print_message(error, syntax_error(0,Between,Msg,Tokens,AfterError))
              ; print_message(error, Excp)
              ),
              %% %% should really just remove those assertaed above
              %% retractall(konsult_context(_,_)),
              erase(R1), erase(R2), erase(R3),
              working_directory(_,WD),
              %% raise_exception(Excp)             % reset module not needed
              raise_exception('konsult aborted')
              
          )
       ),
       seen
   ),
   %% %% should really just remove those assertaed above
   %% retractall(konsult_context(_,_)),
   erase(R1), erase(R2), erase(R3),
   working_directory(_,WD),

   statistics(runtime, [T1|_]),
   %% module might have changed in read_klauses_stream
   module(M),
   konsult_file_segments(FileSegments, M, Act1, Act),
   module(M0),
   statistics(runtime, [T2|_]),
   TR is T1-T0,
   TK is T2-T1,
   format('~N{~w konsulted, ~w msec read ~w msec konsult}~n', [File,  TR, TK]).


read_klauses_stream(File, Dir, FileSegments, M, Act0, Act) :-
   working_directory(_,Dir),
   act_new_file(File, Act0, Act1),
   read_klauses(FileSegments, M, Act1, Act).


%% konsult_file_segments(FileSegments, M, Act0, Act)
konsult_file_segments(FileSegments, M, Act0, Act) :-
   SeenPreds = [],
   konsult_file_segments(FileSegments, M, SeenPreds, Act0, Act).


konsult_file_segments([], _M, _Seen, Act0, Act) :-
   %% 960901 This used to be called at end of a single segment.
   act_file_final(Act0, Act).


konsult_file_segments([Segment|FileSegments], M0, Seen0, Act0, Act) :-
   konsult_klauses(Segment, M0, M, Seen0, Seen1, Act0, Act1),
   konsult_file_segments(FileSegments, M, Seen1, Act1, Act).


konsult_klauses(Klauses, M0, M, Seen0, Seen, Act0, Act) :-
   konsult_klauses(Klauses, 'none', M0, M, Seen0, Seen, Act0, Act).


konsult_klauses([], _Prev, M0, M, Seen0, Seen, Act0, Act) :-
   M = M0,
   Seen = Seen0,
   %% 960901 Was, but it should be done when file is done, not when segment is done
   %% act_file_final(Act0, Act)
   
   Act = Act0.


konsult_klauses([K-Kl|Kls], Prev, M0, M, Seen0, Seen, Act0, Act) :-
   ( K == Prev ->
       Seen1 = Seen0,
       Prev1 = Prev,
       konsult_klause(K, Kl, M0, M1, Act0, Act1)
   ; otherwise ->
       ( memberchk(K, Seen0) ->
           %% 971031 A kludge to allow discontigous clauses
           %%        Call konsult_klaus instead of konsult_new_klause
           %%        when we detect that the clause is for an allready
           %%        seen predicate.
           %%        This may work for source debug but breaks the
           %%        generic-consult API. In either way it should be
           %%        migrated to the new version in the analyzer.
           %% warn('~w appeared again in file, earlier clauses will be ignored', [K]),
           Prev1 = K,
           Seen1 = Seen0,
           konsult_klause(K, Kl, M0, M1, Act0, Act1)
       ; otherwise ->
           Prev1 = K,
           Seen1 = [K|Seen0],
           konsult_new_klause(K, Kl, M0, M1, Act0, Act1)
       )
   ),
   konsult_klauses(Kls, Prev1, M1, M, Seen1, Seen, Act1, Act).

%konsult_klauses([K-Kl|Kls], Prev, M0, M, Seen0, Seen, Act0, Act) :-
%   ( K == Prev ->
%       Seen1 = Seen0,
%       Prev1 = Prev,
%       konsult_klause(K, Kl, M0, M1, Act0, Act1)
%   ; otherwise ->
%       ( memberchk(K, Seen0) ->
%           warn('~w appeared again in file, earlier clauses will be ignored', [K]),
%           Seen1 = Seen0
%       ; otherwise ->
%           Seen1 = [K|Seen0]
%       ),
%           
%       Prev1 = K,
%       konsult_new_klause(K, Kl, M0, M1, Act0, Act1)
%   ),
%   konsult_klauses(Kls, Prev1, M1, M, Seen1, Seen, Act1, Act).

konsult_klauses(What-Data, _Prev, M0, M, Seen0, Seen, Act0, Act) :-
   Seen = Seen0,
   konsult_klauses_directive(What, Data, M0, M, Act0, Act).


%% !!! Should let some act hook have a go at this too
konsult_klauses_directive(command, _Command, M, M, Act, Act).
konsult_klauses_directive(declaration, Declaration, M0, M, Act0, Act) :-
   Act = Act0,
   ( Declaration = module(M1,_) ->
       module(M1),
       M = M1
   ; otherwise ->
       M = M0
   ).


konsult_klause(F/A, HB, M, M, Act0, Act) :-
   act_clause(F/A, HB, M, Act0, Act).


konsult_new_klause(F/A, HB, M, M, Act0, Act) :-
   act_new_clause(F/A, HB, M, Act0, Act).


read_klauses(Klauses, M, Act0, Act) :-
   current_input(Stream),
   current_stream(File, _, Stream),
   read_klauses(M, Stream, File, Klauses, Act0, Act).


read_klauses(M, Stream, File, [KSeg|Klauses], Act0, Act) :-
   read_klauses_segment(M, Stream, File, KSeg, Klauses, Act0, Act).


read_klauses_segment(M, Stream, File, KSeg, Klauses, Act0, Act) :-
   line_count(Stream, L0),
   xread_term(Term, [singletons(Singles), source_positions(SrcPos0)]),
   line_count(Stream, L1),
   warn_singles(Singles, predicate, L0, L1, File), % !!! for now
   SrcPos = File-SrcPos0,
   expand_term(Term, ET),
   ( var(ET) ->                                 % terminate segment
       KSeg = [],                               % should warn
       read_klauses(M, Stream, File, Klauses, Act0, Act)
   ; ET = end_of_file ->
       KSeg = [],
       Klauses = [],
       Act = Act0
   ; ET = [] ->                                 % terminate?
       read_klauses_segment(M, Stream, File, KSeg, Klauses, Act0, Act)
   ; ET = [_|_] ->
       process_clause_list(ET, SrcPos, M, Stream, File,
                                 KSeg, Klauses, Act0, Act)
   ; otherwise ->
       read_klause_segment(ET, SrcPos, M, Stream, File, KSeg, Klauses, Act0, Act)
   ).


%% process_clause_list(Ts, SrcPos, M, Stream, File, KSeg, Klauses, Act0, Act)
process_clause_list([], _SrcPos, M, Stream, File, KSeg, Klauses, Act0, Act) :-
   read_klauses_segment(M, Stream, File, KSeg, Klauses, Act0, Act).

process_clause_list([T|Ts], SrcPos, M, Stream, File, KSeg, Klauses, Act0, Act) :-
   process_clause(T, T, Ts, SrcPos, M, Stream, File, KSeg, Klauses, Act0, Act).
   

%% process_clause_list(Ts, SrcPos, M, Stream, File, Klauses, Act0, Act)
process_clause_list([], _SrcPos, M, Stream, File, Klauses, Act0, Act) :-
   read_klauses(M, Stream, File, Klauses, Act0, Act).

process_clause_list([T|Ts], SrcPos, M, Stream, File, Klauses, Act0, Act) :-
   Klauses = [KSeg|Klauses1],
   process_clause(T, T, Ts, SrcPos, M, Stream, File, KSeg, Klauses1, Act0, Act).
   

read_klause_segment(T, SrcPos, M, Stream, File, KSeg, Klauses, Act0, Act) :-
   process_clause(T, T, [], SrcPos, M, Stream, File, KSeg, Klauses, Act0, Act).


%% process_clause(T, T, Ts, SrcPos, M, Stream, File, KSeg, Klauses, Act0, Act)
process_clause((:-(Declaration)), _T, Ts, SrcPos, M0, Stream, File,
                KSeg, Klauses, Act0, Act) :- !,
       KSeg = [],
   ( handle_declaration(Declaration, M0, M) ->
       Act1 = Act0,
       Klauses = [declaration-Declaration|Klauses1]
   ; act_command_read(Declaration, Act0, Act1) ->
       M = M0,
       Klauses = [command-Declaration|Klauses1]
   ; otherwise ->
       M = M0,
       Act1 = Act0,
       Klauses = [command-Declaration|Klauses1],
       (\+ call(M:Declaration) -> true; true)
   ),
   process_clause_list(Ts, SrcPos, M, Stream, File, Klauses1, Act1, Act).


process_clause((?-(Declaration)), T, Ts, SrcPos, M, Stream, File,
                KSeg, Klauses, Act0, Act) :- !,           
   process_clause((:-(Declaration)), T, Ts, SrcPos, M, Stream, File,
                  KSeg, Klauses, Act0, Act).

process_clause((Head :- Body), T, Ts, SrcPos, M, Stream, File,
                KSeg, Klauses, Act0, Act) :- !,
   head_info(Head, M, FA, MH),
   %% functor(Head, F, A),
   ( MH \== M -> MFA = MH:FA ; MFA = FA ),
   Kl = MFA-(H:-B),
   act_clause_read(Head, Body, T, SrcPos, H, B, Act0, Act1),
   KSeg = [Kl|KSeg1],
   process_clause_list(Ts, SrcPos, M, Stream, File, KSeg1, Klauses, Act1, Act).

process_clause(Head, T, Ts, SrcPos, M, Stream, File,
               KSeg, Klauses, Act0, Act) :-
   head_info(Head, M, FA, MH),
   %% functor(Head, F, A),
   ( MH \== M -> MFA = MH:FA ; MFA = FA ),
   Kl = MFA-(H:-B),
   Body = true,
   act_clause_read(Head, Body, T, SrcPos, H, B, Act0, Act1),
   KSeg = [Kl|KSeg1],
   process_clause_list(Ts, SrcPos, M, Stream, File, KSeg1, Klauses, Act1, Act).


head_info(Head, _M, _FA, _MH) :- var(Head), fail.  % !!! should warn

head_info(M:H, _M, FA, MH) :- !,
   head_info(H, M, FA, MH).

head_info(H, M, F/A, M) :-
   functor(H, F, A).


%% as described in SICStus 3 manual section 5.2
handle_declaration(multifile(_Preds), M, M) :-
   warn('multifile declaration ignored').
handle_declaration(dynamic(_Preds), M, M).
handle_declaration(block(_Specs), M, M) :-
   warn('block declaration ignored').
handle_declaration(meta_predicate(_MetaPredSpecs), M, M) :-
   warn('meta_predicate declaration ignored').
handle_declaration(module(ModuleName,PublicPredicateList), _M, M) :-
   M = ModuleName,
   %% There is no documented (nor a simple) way to export a predicate
   %% from a module. Thus abolishing is not a good idea either as it
   %% breaks up the export/import links.
   %% Doing this right will have to wait until/if this is incorporated
   %% into SICStus
   %%   ( current_module(ModuleName) ->
   %%       ( %% abolish all existing predicates from this module
   %%           current_predicate(_Name, ModuleName:H),
   %%           functor(H,F,A),
   %%           abolish(ModuleName:(F/A)),
   %%           fail
   %%       ; true
   %%       )
   %%   ; otherwise ->                               % new module
   %%       true
   %%   ),
   warn('~w - public predicate list ignored for module ~w',
        [PublicPredicateList, ModuleName]),
   module(ModuleName).
handle_declaration(public(_PredSpecs), M, M).
handle_declaration(mode(_ModeSPecs), M, M).
handle_declaration(parallell(_PredSpecs), M, M) :-
   warn('parallell declaration ignored').
handle_declaration(sequential(_PredSpecs), M, M) :-
   warn('sequential declaration ignored').



%% warn_singles(Singles, FA, L0, L1, Path).
warn_singles([], _FA, _L0, _L1, _Path).
warn_singles([V|Vs], FA, L0, L1, _Path) :-
   (  prolog_flag(single_var_warnings,on) ->
       LStart is L0+1,
       LEnd is L1+1,
       filter_var_list([V|Vs], VarNames),
       %% We CAN use use our portray_message hook as it
       %% uses prolog_load_kontext or konsult_load_context
       print_message(warning, singletons(VarNames,FA,LStart,LEnd))
       %% Prefix = '{Warning: ', Suffix = '}',
       %% format(user_error,
       %%        '~a~q - singleton variables in ~q in lines ~d-~d file \'~w\' ~a~n',
       %%        [Prefix,VarNames,FA,LStart,LEnd,Path,Suffix])

   ; otherwise ->
       true
   ).

filter_var_list([], []).

filter_var_list([Name=_|NVs], ['$VAR'(Name)|Names]) :-
   filter_var_list(NVs, Names).


warn(FormatString, Args) :-
   format('~N{Warning: ',[]),
   format(FormatString, Args),
   write('}'),
   nl.

warn(X) :-
   format('~N{Warning: ~a}~n', [X]).




%% This should be a builtin!
file_directory(File, Dir) :-
   %% prolog:absolute_file_name(Name, Suffix, Goal, AbsFile, AbsDir, Found)
   prolog:absolute_file_name(File, '', file_directory, _, AbsDir, _),
   Dir = AbsDir.
%% Hacked by Per Mildner 1995 to support source debugging
%%        Atom tokens used to be atom(+Atom), changing to
%%        atom(+Atom,+Pos,-Term) (actually Pos can be var for unknown)
%% (ie token/term level file position info)
%% Internal stuff I will use:
%% prolog:'$bdd_equiv'(A,B) (a.k.a. eq) to track terms through term_expansion
%% prolog:'$stream_count'(+Stream, -CharCount, -LineCount, -LinePos), to track file position
%% Sketch:
%% Add an option to read_term source_positions(X)
%% Upon return X is a list of Term-Pos entries where Term are
%% non-atomic subterms of the read term.
%% As multiple subterms may be == the intention is that the caller
%% will look up sub-terms with eq. (This will also correctly track
%% terms through term_expansion (and goal_expansion))
%%
%% To implement this we need a new read_tokens that attach positions
%% to the tokens. There are sevaral ways to do this but the one used
%% here is to attach it to atom tokens.
%%
%% I believe this could be incorporated into the real read with no
%% performance penalty. The major difference is that atom tokens goes
%% from arity one to arity three and an extra arg to some
%% predicates. If it is, then prolog:'$read_token' should be modified
%% to provide the new-style atom tokens directly
%%
%% for now pos will be line-numbers. We really would like this to come
%% back from '$read_token'/5 but until that time we will have to use
%% the line-position AFTER the token was read (usually better than
%% before token was read as read_token skips preceeding whitespace,
%% eg, comments)
%% 
%% Look for [PM] and prednames starting with x

%% Indentation got lost somewhere, sorry about that :-)



%   File   : READ.PL
%   Author : D.H.D.Warren + Richard O'Keefe
%   Updated: 5 July 1984
%   Purpose: Read Prolog terms in Dec-10 syntax.

/*  Adapted from shared code written by the same authors; all changes */
/*  Copyright (C) 1988, Swedish Institute of Computer Science. */


/*
    Modified by Alan Mycroft to regularise the functor modes.
    This is both easier to understand (there are no more '?'s),
    and also fixes bugs concerning the curious interaction of cut with
    the state of parameter instantiation.

    Since this file doesn't provide "metaread", it is considerably
    simplified.  The token list format has been changed somewhat, see
    the comments in the RDTOK file.

    I have added the rule X(...) -> apply(X,[...]) for Alan Mycroft.
*/


%% :- public read/1, read/2, read_term/2, read_term/3.

/*
xread(Stream, X) :-
   prolog:'$syntax'(EH, EH),
   read_top_level(Stream, Ans, _, read(tream,X), EH, S, S),
   read_answer(Ans, X).

xread(X) :-
   prolog:'$input'(Stream, Stream),
   prolog:'$syntax'(EH, EH),
   read_top_level(Stream, Ans, _, read(X), EH, S, S),
   read_answer(Ans, X).
*/


xread_term(Term, Options) :-
   prolog:'$input'(Stream, Stream),
   xread_term(Stream, Term, Options).

xread_term(Stream, Term, Options) :-
   prolog:'$syntax'(EH0, EH0),
        %% [PM] added style arg SrcPos
   read_term_options(Options, style(EH0,'$ignore','$ignore',false,'$ignore'),
                              style(EH,VarNames,Singletons,Cycles,SrcPos)), !,
   read_top_level(Stream, Ans, Variables, 
                  read_term(Stream,Term,Options), EH, S, S),
   read_answer(Ans, X, Tokens),
   (   VarNames == '$ignore' -> true
   ;   read_term_variables(Variables, VarNames, [])
   ),
   (   Singletons == '$ignore' -> true
   ;   read_term_singletons(Variables, Singletons, [])
   ),
   (   SrcPos == '$ignore' -> true
   ;   read_term_source_pos(Tokens, SrcPos)     % [PM]
   ),
   (   Cycles=false -> Term = X
   ;   read_term_cycles(X, Term)
   ).
xread_term(Stream, Term, Options) :-
   illarg(domain(term,read_term_options), read_term(Stream,Term,Options), 3).


read_term_options(-, _, _) :- !, fail.
read_term_options([], Style, Style).
read_term_options([X|Xs], Style0, Style) :-
   read_term_option(X, Style0, Style1),
   read_term_options(Xs, Style1, Style).

read_term_option(-, _, _) :- !, fail.
read_term_option(syntax_errors(Q), style(_,B,C,D,E), style(Q,B,C,D,E)) :-
   syntax_error_option(Q).
read_term_option(variable_names(Q), style(A,_,C,D,E), style(A,Q,C,D,E)).
read_term_option(singletons(Q), style(A,B,_,D,E), style(A,B,Q,D,E)).
read_term_option(cycles(Q), style(A,B,C,_,E), style(A,B,C,Q,E)) :-
   bool_option(Q).
read_term_option(source_positions(Q), style(A,B,C,D,_), style(A,B,C,D,Q)). % [PM]


syntax_error_option(-) :- !, fail.
syntax_error_option(error).
syntax_error_option(dec10).
syntax_error_option(quiet).
syntax_error_option(fail).

read_term_variables([]) --> [].
read_term_variables([String-Var|R]) -->
   [Name=Var],
   {atom_chars(Name, String)},
   {read_term_variables_skip(R, String, R1)},
   read_term_variables(R1).

read_term_variables_skip([String0-_|R0], String, R) :-
   String==String0, !,
   read_term_variables_skip(R0, String, R).
read_term_variables_skip(R, _, R).

read_term_singletons([]) --> [].
read_term_singletons([String-Var|R]) -->
   {read_term_variables_skip(R, String, R1)},
   {String = [C|_]},
   (   {R==R1, C=\="_"} -> [Name=Var], {atom_chars(Name, String)}
   ;   []
   ),
   read_term_singletons(R1).


read_term_cycles(X, Y) :- var(X), !, Y = X.
read_term_cycles(X, Y) :- atomic(X), !, Y = X.
read_term_cycles(@(X,List), X) :- !,
   read_term_bind(List).
read_term_cycles(X, Y) :-
   functor(X, F, N),
   functor(Y, F, N),
   read_term_cycles(0, N, X, Y).

read_term_cycles(N, N, _, _) :- !.
read_term_cycles(A, N, X, Y) :-
   B is A+1,
   arg(B, X, Arg),
   arg(B, Y, Arg1),
   read_term_cycles(Arg, Arg1),
   read_term_cycles(B, N, X, Y).

read_term_bind([]).
read_term_bind([X=X|L]) :- read_term_bind(L).



%% Filter out entries atom(Atom,Pos,Term) where Pos is nonvar (an int)
%% and Term is a struct
%% NOTE: Currently we do not remove duplicate (ie no real sort)
%% Consider making this a binary tree, this would probably become
%% rather balanced. For this we would need pointer comparisons
%% though... We could also use an ordinary compare based tree with
%% lists for overflow buckets.
%% read_term_source_pos(Tokens, SrcPos)
read_term_source_pos([], []).
read_term_source_pos([T|Ts], SrcPos) :-
   read_term_source_pos(T, Ts, SrcPos).

read_term_source_pos(atom(_Atom,Pos,Term), Ts, SrcPos) :-
   nonvar(Term),
   compound(Term), !,                          % arity >= 1
   SrcPos = [pos(Pos,Term)|SrcPos1],
   read_term_source_pos(Ts, SrcPos1).
read_term_source_pos(_, Ts, SrcPos) :-
   read_term_source_pos(Ts, SrcPos).


read_top_level(Stream, Answer, Variables) :-
   read_top_level(Stream, Ans, Variables, 0, dec10, S, S),
   read_answer(Ans, Answer).

read_with_prompt(Stream, Answer, Variables, T) :-
   read_with_prompt(T, Prompt),
   prolog:'$prompt'(Old, Prompt),
   read_top_level(Stream, Ans, Variables, 0, dec10, [atom(T,_,_)|S], S),
   prolog:'$prompt'(_, Old),
   read_answer(Ans, Answer).

read_with_prompt(?-, '| ?- ').
read_with_prompt(:-, '| :- ').

read_top_level(Stream, Ans, Variables, Goal, EH, Tokens0, Tokens) :-
        prolog:'$input'(Curr, Stream), !,
        repeat,
           prolog:'$stream_count'(Stream, _, AL, _),
           AL1 is AL+1,
           %% prolog:internal_catch(source_debug:xread_tokens(Tokens,Variables), Excp, true),
           on_exception(Excp,
              xread_tokens(Tokens,Variables),
              true
           ),
           prolog:'$stream_count'(Stream, _, BL, _),
           BL1 is BL+1,
           (   nonvar(Excp) -> Ans = exception(Excp)
       ;   read(Tokens0, 1200, Term, LeftOver),
           all_read(LeftOver) ->
           Ans = answer(Term, Tokens),          % [PM], added Tokens
           prolog:purge_internal('$syntax error'(_,_)) % erase any leftovers
       ;   syntax_exception(Tokens0, AL1, BL1, Goal, SynExcp),
        (   EH=dec10 -> print_error(SynExcp), fail
        ;   EH=quiet -> Ans = fail
        ;   EH=fail -> Ans = fail, print_error(SynExcp)
        ;   /* EH=error -> */ Ans = exception(SynExcp)
        )
       ), !,
        prolog:'$input'(_, Curr).
read_top_level(Stream, _, _, Goal, _, _, _) :-
   prolog:'$cerror'(Goal, 1, Stream).

read_answer(answer(Answer, _Tokens), Answer).
read_answer(exception(Excp), _) :- raise_exception(Excp).

%% [PM], return token list with source positions too.
read_answer(answer(Answer, Tokens), Answer, Tokens).
read_answer(exception(Excp), _, _) :- raise_exception(Excp).


%   all_read(+Tokens)
%   checks that there are no unparsed tokens left over.

all_read([.]) :- !.
all_read(S) :-
   syntax_error([operator,expected,after,expression], S).


%   expect(Token, TokensIn, TokensOut)
%   reads the next token, checking that it is the one expected, and
%   giving an error message if it is not.  It is used to look for
%   right brackets of various sorts, as they're all we can be sure of.

expect(Token, [Token|Rest], Out) :- !, Out=Rest.
expect(Token, S0, _) :-
   syntax_error([Token,or,operator,expected], S0).


%   I want to experiment with having the operator information held as
%   ordinary Prolog facts.  For the moment the following predicates
%   remain as interfaces to current_op.
%   current_prefixop(O -> Self, Rarg)
%   current_postfixop(O -> Larg, Self)
%   current_infixop(O -> Larg, Self, Rarg)

%   after_prefix_op(+Op, +Prec, +ArgPrec, +Rest, +Precedence, -Ans, -LeftOver)
% [PM] added Term arg
after_prefix_op(Op, Oprec, Aprec, S1, Precedence, Answer, Term, S) :-
   S1 = [Token|_],
   \+ cant_start_expr(Token),
   (   Precedence < Oprec ->
       syntax_error([prefix,operator,Op,in,context,with,precedence,
                     Precedence],
               S1)
   ;   read(S1, Aprec, Arg, S2),
       functor(Term, Op, 1),
       arg(1, Term, Arg),
       read_rest(S2, Oprec, Term, Precedence, Answer, S)
   ).


cant_start_expr(')').
cant_start_expr(rbrack(_)).  % ']'
cant_start_expr('}').
cant_start_expr('|').
cant_start_expr(',').
cant_start_expr(.).


%   read(+TokenList, +Precedence, -Term, -LeftOver)
%   parses a Token List in a context of given Precedence,
%   returning a Term and the unread Left Over tokens.

read([], _, _, _) :-
   syntax_error([expression,expected], []).
read([Token|RestTokens], Precedence, Term, LeftOver) :-
   read(Token, RestTokens, Precedence, Term, LeftOver).


%   read(+Token, +RestTokens, +Precedence, -Term, -LeftOver)
% [PM] modified for atom(Atom,Pos,Term)
read(X, _, _, _, _) :- var(X), !, fail.      % space saver
read(var(_,_,_,Variable), S0, Precedence, Answer, S) :- !,
   read_rest(S0, 0, Variable, Precedence, Answer, S).
read(atom(-,_,_), [Token|S1], Precedence, Answer, S) :-
   number_token(Token, Number), !,
   Negative is -Number,
   read_rest(S1, 0, Negative, Precedence, Answer, S).
read(atom(+,_,_), [Token|S1], Precedence, Answer, S) :-
   number_token(Token, Number), !,
   Negative is +Number,
   read_rest(S1, 0, Negative, Precedence, Answer, S).
read(atom(Functor,_Pos,Term), ['('|S1], Precedence, Answer, S) :- !,
   read(S1, 999, Arg1, S2),
   read_args(S2, RestArgs, S3),
   Term =.. [Functor,Arg1|RestArgs], !,
   read_rest(S3, 0, Term, Precedence, Answer, S).
read(atom(Functor,_Pos,Term), S0, Precedence, Answer, S) :-
   prolog:current_prefixop(Functor, Prec, Right),
   after_prefix_op(Functor, Prec, Right, S0, Precedence, Answer, Term, S).
read(atom(Atom,_,Term), S0, Precedence, Answer, S) :- !,
   Term = Atom,
   read_rest(S0, 0, Atom, Precedence, Answer, S).
read(number(Number), S0, Precedence, Answer, S) :-
   number(Number), !,
   read_rest(S0, 0, Number, Precedence, Answer, S).
read(lbrack(_), [rbrack(_)|S1], Precedence, Answer, S) :- !,
   read(atom([],_,_), S1, Precedence, Answer, S).
read(lbrack(_), S1, Precedence, Answer, S) :- !,
   read(S1, 999, Arg1, S2),
   read_list(S2, RestArgs, S3), !,
   read_rest(S3, 0, [Arg1|RestArgs], Precedence, Answer, S).
read('(', S1, Precedence, Answer, S) :- !,
   read(S1, 1200, Term, S2),
   expect(')', S2, S3), !,
   read_rest(S3, 0, Term, Precedence, Answer, S).
read(' (', S1, Precedence, Answer, S) :- !,
   read(S1, 1200, Term, S2),
   expect(')', S2, S3), !,
   read_rest(S3, 0, Term, Precedence, Answer, S).
read('{', ['}'|S1], Precedence, Answer, S) :- !,
   read(atom({},_,_), S1, Precedence, Answer, S).
read('{', S1, Precedence, Answer, S) :- !,
   read(S1, 1200, Term, S2),
   expect('}', S2, S3), !,
   read_rest(S3, 0, {Term}, Precedence, Answer, S).
read(string(List), S0, Precedence, Answer, S) :- !,
   read_rest(S0, 0, List, Precedence, Answer, S).
read(Token, S0, _, _, _) :-
   syntax_error([Token,cannot,start,an,expression], S0).

number_token(number(N), N) :- number(N).
number_token(atom(inf,_,_), inf).
number_token(atom(nan,_,_), nan).


%   read_rest(+Tokens, +LPrec, +Term, +Prec, -Answer, -LeftOver)
%   is called by read/4 after it has read a primary (the Term).  It
%   checks for following postfix or infix operators.  The tokens
%   Tokens-LeftOver are parsed and combined with Term into Answer;
%   LPrec and Prec are the lower and upper bounds for the precedence
%   of the topmost operator.
% [PM] modified for atom(Atom,Pos,Term)
read_rest([atom(F,_,Expr)|S1], LPrec, Term, Precedence, Answer, S) :-
   prolog:current_infixop(F, L, O, R),
   Precedence >= O, LPrec =< L,
   (   prolog:current_postfixop(F, _, _) -> true
   ;   !
   ),
   read(S1, R, Other, S2),
   functor(Expr, F, 2),
   arg(1, Expr, Term),
   arg(2, Expr, Other),
   read_rest(S2, O, Expr, Precedence, Answer, S).
read_rest([atom(F,_,Expr)|S1], LPrec, Term, Precedence, Answer, S) :-
   prolog:current_postfixop(F, L, O),
   Precedence >= O, LPrec =< L,
   (   prolog:current_infixop(F, _, _, _) -> true
   ;   !
   ),
   functor(Expr, F, 1),
   arg(1, Expr, Term),
   read_rest(S1, O, Expr, Precedence, Answer, S).
read_rest([','|S1], LPrec, Term, Precedence, Answer, S) :-
   Precedence >= 1000, LPrec < 1000, !,
   read(S1, 1000, Next, S2),
   read_rest(S2, 1000, (Term,Next), Precedence, Answer, S).
read_rest(['|'|S1], LPrec, Term, Precedence, Answer, S) :-
   Precedence >= 1100, LPrec < 1100, !,
   read(S1, 1100, Next, S2),
   read_rest(S2, 1100, (Term;Next), Precedence, Answer, S).
read_rest(S, _, Term, _, Term, S).


%   This business of syntax errors is tricky.  When an error is detected,
%   we have to write out a message.  We also have to note how far it was
%   to the end of the input, and for this we are obliged to use the data-
%   base.  Then we fail all the way back to read(), and that prints the
%   input list with a marker where the error was noticed.  If subgoal_of
%   were available in compiled code we could use that to find the input
%   list without hacking the data base.  The really hairy thing is that
%   the original code noted a possible error and backtracked on, so that
%   what looked at first sight like an error sometimes turned out to be
%   a wrong decision by the parser.  This version of the parser makes
%   fewer wrong decisions, and my goal was to get it to do no backtracking
%   at all.  This goal has not yet been met, and it will still occasionally
%   report an error message and then decide that it is happy with the input
%   after all.  Sorry about that.

:- dynamic
   '$syntax error'/2.

syntax_error(Message, List) :-
   length(List, Length),
   prolog:asserta_internal('$syntax error'(Message,Length), _), !,
   fail.

%   read_args(+Tokens, -TermList, -LeftOver)
%   parses {',' expr(999)} ')' and returns a list of terms.

read_args([','|S1], TermList, S) :- !,
   TermList = [Term|Rest],
   read(S1, 999, Term, S2), /*!, MC*/
   read_args(S2, Rest, S).
read_args([')'|S1], TermList, S) :- !,
   TermList = [],
   S = S1.
read_args(S, _, _) :-
   syntax_error([', or )',expected,in,arguments], S).


%   read_list(+Tokens, -TermList, -LeftOver)
%   parses {',' expr(999)} ['|' expr(999)] ']' and returns a list of terms.

read_list([','|S1], TermList, S) :- !,
   TermList = [Term|Rest],
   read(S1, 999, Term, S2), /*!, MC*/
   read_list(S2, Rest, S).
read_list(['|'|S1], Rest, S) :- !,
   read(S1, 999, Rest, S2), /*!, MC*/
   expect(rbrack(_), S2, S).
read_list([rbrack(_)|S1], TermList, S) :- !,
   TermList = [],
   S = S1.
read_list(S, _, _) :-
   syntax_error([', | or ]',expected,in,list], S).

syntax_exception(List0, AL, BL, Goal,
            syntax_error(Goal,between(AL,BL),Msg,List,AfterError)) :-
   trim_token_list(List0, List),                % [PM] change atom/3 to atom/1
   the_syntax_error([], 1000000, Msg, AfterError).

%% [PM]
trim_token_list([], []).
trim_token_list([T|Ts], [TT|TTs]) :-
   ( T = atom(A,_,_) ->
       TT = atom(A)
   ; otherwise ->
       TT = T
   ),
   trim_token_list(Ts, TTs).


the_syntax_error(Msg0, AfterError0, Msg, AfterError) :-
   prolog:clause_internal('$syntax error'(Msg1,AfterError1), Ptr), !,
   prolog:'$erase'(Ptr),
   (   AfterError0 > AfterError1 ->
         the_syntax_error(Msg1, AfterError1, Msg, AfterError)
   ;   /* true */
         the_syntax_error(Msg0, AfterError0, Msg, AfterError)
   ).
the_syntax_error(Msg, AfterError, Msg, AfterError).


/* [PM] original
read_tokens(TokenList, Variables) :-
   get(NextCh),
   prolog:'$prompt'(Old, '     '),
   read_tokens(NextCh, 0, Tokens, Variables0),
   prolog:'$prompt'(_, Old),
   sort(Variables0, Variables1),
   tokens_peep(Variables1, Variables),
   (   Tokens = [] ->         % genuine EOF
       TokenList = [atom(end_of_file),.],
       get0(_)          % read past ungotten EOF char
   ;   TokenList = Tokens
   ).
*/


xread_tokens(TokenList, Variables) :-
   get(NextCh),
   prolog:'$prompt'(Old, '     '),
   xread_tokens(NextCh, 0, Tokens, Variables0),
   prolog:'$prompt'(_, Old),
   sort(Variables0, Variables1),
   tokens_peep(Variables1, Variables),
   (   Tokens = [] ->         % genuine EOF
       TokenList = [atom(end_of_file,_,_),.],
       get0(_)          % read past ungotten EOF char
   ;   TokenList = Tokens
   ).

xsource_pos(Pos) :-
   prolog:'$input'(Stream, Stream),
   %% prolog:'$stream_count'(+Stream, -CharCount, -LineCount, -LinePos)
   prolog:'$stream_count'(Stream, _CharCount, LineCount, LinePos),
   ( LinePos == 0 ->
       %% This fixes a problem when the token ends at end-of-line, a
       %% common case for '->'. This makes the kludge for making
       %% conditions in if-then-else (eg, otherwise ->) work.
       Pos is LineCount-1
   ; otherwise ->
       Pos = LineCount
   ).

%% [PM] What we would like $read_token to do
read_token(Ch, State, Token, NextCh, NextState) :-
   prolog:'$read_token'(Ch, State, Token0, NextCh0, NextState0),
   ( Token0 = atom(Atom) ->
       xsource_pos(Pos),
       Token = atom(Atom,Pos,_Term)
   ; (Token0 = var(MondoCode,ListofAscii,RefVar) ->
       xsource_pos(Pos),
       Token = var(MondoCode,Pos,ListofAscii,RefVar)
     ; (Token0 = '[' ->
         xsource_pos(Pos),
         Token = lbrack(Pos)
       ; (Token0 = ']' ->
           xsource_pos(Pos),
           Token = rbrack(Pos)
         ; otherwise ->
             Token = Token0
         )
       )
     )
   ),
   %% !! DONT CHANGE THIS! $read_token requires dereferenced unit var
   %% as args 3,4 and 5 (ie X2, X3 and X4) (Note, this applies to
   %% Token0 above too)
   NextCh=NextCh0,
   NextState=NextState0.

%% [PM]
xread_tokens(Ch, State, Tokens, Variables) :-
   read_token(Ch, State, Token, NextCh, NextState),
   xread_tokens(Token, NextCh, NextState, Tokens, Variables).

xread_tokens([], Ch, _, Tokens, Variables) :- !,
   Tokens = [],
   Variables = [],
   prolog:'$unget'(Ch).          % unget EOF char (see above)
xread_tokens(.,  NextCh, _, Tokens, Variables) :- !,
   Tokens = [.],
   Variables = [],
   after_full_stop(NextCh).
xread_tokens(Token, NextCh, NextState, Tokens0, Variables0) :-
   Token = var(_,_,S,_),
   S \== "_", !,
   Tokens0 = [Token|Tokens],
   Variables0 = [Token|Variables],
   xread_tokens(NextCh, NextState, Tokens, Variables).
xread_tokens(Token, NextCh, NextState, [Token|Tokens], Variables) :-
   xread_tokens(NextCh, NextState, Tokens, Variables).


/* [PM] original
read_tokens(Ch, State, Tokens, Variables) :-
   prolog:'$read_token'(Ch, State, Token, NextCh, NextState),
   read_tokens(Token, NextCh, NextState, Tokens, Variables).

read_tokens([], Ch, _, Tokens, Variables) :- !,
   Tokens = [],
   Variables = [],
   prolog:'$unget'(Ch).          % unget EOF char (see above)
read_tokens(.,  NextCh, _, Tokens, Variables) :- !,
   Tokens = [.],
   Variables = [],
   after_full_stop(NextCh).
read_tokens(Token, NextCh, NextState, Tokens0, Variables0) :-
   Token = var(_,_,S,_),
   S \== "_", !,
   Tokens0 = [Token|Tokens],
   Variables0 = [Token|Variables],
   read_tokens(NextCh, NextState, Tokens, Variables).
read_tokens(Token, NextCh, NextState, [Token|Tokens], Variables) :-
   read_tokens(NextCh, NextState, Tokens, Variables).

*/


% cleaning up after full stop
after_full_stop(0'%) :- !, prolog:'$skip'(10).
after_full_stop(-1) :- !, prolog:'$unget'(-1).
after_full_stop(_).

tokens_peep([], []).
tokens_peep([X|R], [S-V|Vs]) :-
   X = var(Code,_,S,V),
   (   R = [var(Code,_,S,V)|_]
   ->  tokens_peep(R, Vs)
   ;   tokens_peep(R, Vs)
   ).

%%% Copyright (c) 1995 Per Mildner, all rights reserved.
%%% email: Per.Mildner@csd.uu.se
%% 970509 N+1:th change in how line-numbers are handled.
%%        Now use indirection to allow fast breakpoints.
%% 960902 . added a kludge to enable indexing for code like f(X,...):- X=...

:- use_module(library(lists)).
:- use_module(library(ordsets)).
%% ???
:- dynamic 'line mapping'/2.
%%
:- dynamic 'pos marker goal'/1.
%% 970509
:- dynamic 'pos marker table goal'/2.           % pos-marker-goal->table-goal

act_init('konsult'([])).

%% Should comply with pos_goal_head/1 and pos_goal_table_head/1!
pos_pred_heads(Path, PosPredHead, PosTablePredHead) :-
   src_pos_goal1(_, Path, PosPredHead),
   src_pos_table_goal1(_, Path, PosTablePredHead).


src_pos_goal(Line, Path, user:PosGoal) :-
   src_pos_goal1(Line, Path, PosGoal).

src_pos_table_goal(Line, Path, user:PosTablePredHead) :-
   src_pos_table_goal1(Line, Path, PosTablePredHead).



src_pos_goal1(Line, Path, PosGoal) :-
   functor(PosGoal, Path, 1),
   arg(1,PosGoal,Line).

src_pos_table_goal1(Line, Path, PosTablePredHead) :-
   atom_chars(Path,PathChars),
   append(PathChars, "//", TableNameChars),
   atom_chars(TableName, TableNameChars),
   functor(PosTablePredHead, TableName, 1),
   arg(1,PosTablePredHead,Line).

%% Used from profile to recover th path from TablePred Name/Arity
pos_table_path(Name/1, Path) :-
   atom_chars(Name, Chars),
   append(PathChars, "//", Chars), !,
   atom_chars(Path, PathChars).


%% 970509 abolish a pred user:Name/Arity given a goal Name(....)
%%        Check first that the pred exists to avoid warnings.
safe_abolish_head(Goal0) :-
   functor(Goal0,Name,Arity),
   functor(Goal,Name,Arity),
   current_predicate(Name, user:Goal), !,
   abolish(user:Name/Arity).
safe_abolish_head(_Goal).


%% Used from non_konsult_hook when consulting/compiling/loading
%% And at start of konsulting file (act_new_file)
remove_source_debug_info(Path) :-
   reset_profile_data(Path),
   emacs:abolish_line_conditions(Path),
   pos_pred_heads(Path, PosPredHead, PosTablePredHead),
   safe_abolish_head(PosPredHead),
   safe_abolish_head(PosTablePredHead),
   retractall('line mapping'(PosTablePredHead, _ClauseNo)).

       

act_new_file(Path, 'konsult'(Info0), 'konsult'(Info)) :-
   lookup_info_swap(lines, _Default, _Old, X-X, Info0, Info1),
   lookup_info_swap(path, _, _, Path, Info1, Info),
   %% 970612 break out for consult/compile/load
   remove_source_debug_info(Path).
   %% reset_profile_data(Path),
   %% %% 970505, 970509
   %% emacs:abolish_line_conditions(Path),
   %% %% abolish(user:'$ pos'/2)
   %% %% functor(PosPredHead, Path, 1),
   %% pos_pred_heads(Path, PosPredHead, PosTablePredHead),
   %% safe_abolish_head(PosPredHead),
   %% safe_abolish_head(PosTablePredHead),
   %% %% ( current_predicate(Path, user:PosPredHead) ->
   %% %%     abolish(user:Path/1)
   %% %% ; otherwise ->
   %% %%     true
   %% %% ),
   %% retractall('line mapping'(PosTablePredHead, _ClauseNo)).
   

act_clause(_FA, HB, M, 'konsult'(Info), 'konsult'(Info)) :-
   assertz(M:HB).


act_new_clause(FA, HB, M, 'konsult'(Info), 'konsult'(Info)) :-
   FA = F/A, functor(H, F, A),
   ( current_predicate(_, M:H) ->
       %% Unfortunately abolish removes spy-points as well,
       %% on the other hand, retractall only works for dynamic code
       %% We could either check for dynamic code or
       %% save away the SICStus internal spy info and restore it later.
       spy_info(F, A, M, SpyInfo),
       abolish(M:FA),
       set_spy_info(F, A, M, SpyInfo),                   % restore spying
       retractall(prolog:'$source file'(H, M, _Path))
   ; otherwise ->
       true
   ),
   ( lookup_info(path, Path, Info) ->
        assert(prolog:'$source file'(H, M, Path))
   ; otherwise ->
        true
   ),
   assertz(M:HB).
  

spy_info(N, A, M, Info) :-
   functor(F, N, A),
   ( prolog:'$spypoint'(F, M, on, on) ->
       findall(Cond,
               (
               prolog:'$spypoint_condition'(F,M,X,Y),
               Cond = prolog:'$spypoint_condition'(F,M,X,Y)
               ),
               Conds),
     Info = on(Conds)
   ; otherwise ->
       Info = off
   ).


%% Set spy if unspyed
set_spy_info(N, A, M, Info) :-
   functor(F, N, A),
   ( prolog:'$spypoint'(F, M, on, on) ->
       true                                % do nothing if allready spy
   ; Info = on(Conds) ->
       prolog:'$spypoint'(F, M, off, on),
       ( member(C, Conds),
         asserta(C),
         fail
       ; true
       )
   ; otherwise ->
       Info = off
   ).

act_clause_read(Head, Body, _T, SrcPos, H, B, 'konsult'(Info0), 'konsult'(Info)) :-
   %% Was H = Head,
   %% 960902 The compiler can do first-argument indexing for goals like:
   %%        foo(X,Y,Z) :- X = f(...), ...
   %%        The interpreter can only do first-argument indexing if
   %%        the key is actually at the first position in the head.
   %%        This little trick here transforms a clause:
   %%        foo(X,Y,Z) :- X = f(...), ...
   %%        into
   %%        foo(f(...),Y,Z) :- X = f(...), ...
   %%        There is really no restrictions on the other args (ie no
   %%        need for a linear head or some such).
   ( ( compound(Head), arg(1, Head, FirstParameter), var(FirstParameter),
       nonvar(Body), Body = (FirstGoal,_),
       nonvar(FirstGoal), FirstGoal = (FirstVar = Something),
       FirstVar == FirstParameter, nonvar(Something) ) ->
       Head =.. [F,_FirstParameter|OtherArgs],
       H =.. [F,Something|OtherArgs]
   ; otherwise ->
       H = Head
   ),

   B = KBody,
   src_translate_body(Body, SrcPos, KBody0, ClauseLines, ClTail0),
   %% 960221, only add pos for head if body does not start with a pos
   %% This is for speed to prevent two consecutive pos goals.
   SrcPos = Path-_,
   functor(PosGoal, Path, 1),
   %% 970821 beware of binding if first goal is a variable, eg foo(_):-X.
   ( ( nonvar(KBody0),
       KBody0 = (user:PosGoal,_) ) ->
       KBody = KBody0,
       ClTail0 = ClTail
   ; ( nonvar(KBody0),
       KBody0 = (A,_),
       nonvar(A),
       A = (user:PosGoal,_)) ->
       KBody = KBody0,
       ClTail0 = ClTail
   ; otherwise ->
       src_pos_wrap(Head, SrcPos, KBody, KBody0, ClTail0, ClTail)
   ),

   Old = OldHead-OldTail,
   OldTail = ClauseLines,
   New = OldHead-ClTail,
   source_debug:lookup_info_swap(lines, X-X, Old, New, Info0, Info).

/* to 970820
act_clause_read(Head, Body, _T, SrcPos, H, B, 'konsult'(Info0), 'konsult'(Info)) :-
   %% Was H = Head,
   %% 960902 The compiler can do first-argument indexing for goals like:
   %%        foo(X,Y,Z) :- X = f(...), ...
   %%        The interpreter can only do first-argument indexing if
   %%        the key is actually at the first position in the head.
   %%        This little trick here transforms a clause:
   %%        foo(X,Y,Z) :- X = f(...), ...
   %%        into
   %%        foo(f(...),Y,Z) :- X = f(...), ...
   %%        There is really no restrictions on the other args (ie no
   %%        need for a linear head or some such).
   ( ( compound(Head), arg(1, Head, FirstParameter), var(FirstParameter),
       nonvar(Body), Body = (FirstGoal,_),
       nonvar(FirstGoal), FirstGoal = (FirstVar = Something),
       FirstVar == FirstParameter, nonvar(Something) ) ->
       Head =.. [F,_FirstParameter|OtherArgs],
       H =.. [F,Something|OtherArgs]
   ; otherwise ->
       H = Head
   ),

   B = KBody,
   src_translate_body(Body, SrcPos, KBody0, ClauseLines, ClTail0),
   %% 960221, only add pos for head if body does not start with a pos
   %% This is for speed to prevent two consecutive pos goals.
   SrcPos = Path-_,
   functor(PosGoal, Path, 1),
   ( KBody0 = (user:PosGoal,_) ->
       KBody = KBody0,
       ClTail0 = ClTail
   ; KBody0 = ((user:PosGoal,_),_) ->
       KBody = KBody0,
       ClTail0 = ClTail
   ; otherwise ->
       src_pos_wrap(Head, SrcPos, KBody, KBody0, ClTail0, ClTail)
   ),

   Old = OldHead-OldTail,
   OldTail = ClauseLines,
   New = OldHead-ClTail,
   lookup_info_swap(lines, X-X, Old, New, Info0, Info).
*/


act_command_read(_Command, _Act0, _Act1) :-
   %% fail means do default processing
   fail.


act_command(_Command, _M0, _M1, _Act0, _Act1) :-
   %% fail means do default processing
   fail.

%% 960901 should document if this is really final action on read or
%%        final action on define (it should be final on define file)
act_file_final('konsult'(Info), 'konsult'(Info)) :-
   ( ( lookup_info(path, Path, Info),
       lookup_info(lines, Lines-LinesTail, Info) ) ->
       LinesTail = [],
       list_to_ord_set(Lines,LinesSet),
       define_line_pred(Path, LinesSet),
       %% 970612 Cannot be in konsult_hook in sictus_emacs_debug.pl
       %%        because define_line_preds isn't done then.
       %% 970508 signal breakable lines to emacs
       %%        Should be moved from here to message_handler and do
       %%        it for consult/compile too.
       emacs:emacs_file_breakpoint_info1(Path)
   ; otherwise ->
       true
   ).


%% Define the line indircetion with no breakpoint checks.
define_default_line_indirection(Path) :-
   pos_pred_heads(Path, PosPredHead, PosTablePredHead),
   safe_abolish_head(PosPredHead),
   %% Link the argument
   arg(1,PosPredHead,LineNoArg),
   arg(1,PosTablePredHead,LineNoArg),
   compile_compact_clause(user:(PosPredHead :- PosTablePredHead)).


%% When there (possibly) are breakpoints
define_line_indirection_breaks(Path, Breaks) :-
   pos_pred_heads(Path, PosPredHead, PosTablePredHeadSkel),
   safe_abolish_head(PosPredHead),
   %% define_default_line_indirection(Path),
   ( member(Line, Breaks),
     %% need fresh copy!
     copy_term(PosTablePredHeadSkel, PosTablePredHead),
     pos_pred_heads(Path, PosPredHead, PosTablePredHead),
     arg(1,PosPredHead,LineNoArg),
     arg(1,PosTablePredHead,LineNoArg),
     
     compile_compact_clause(user:(
                             PosPredHead :- LineNoArg == Line, !,
                                PosTablePredHead, % only for profiling
                                %% start trace and make emacs show line
                                emacs:do_pos_marker_for_break(Line, Path)
                                 )
                           ),
     fail
   ; true
   ),
   %% Default clause is same as define_default_line_indirection does
   arg(1,PosPredHead,LineNoArg),
   arg(1,PosTablePredHeadSkel,LineNoArg),
   compile_compact_clause(user:(PosPredHead :- PosTablePredHeadSkel)).

   

define_line_pred(Path, LinesSet) :-
   line_pred_dummy_file(DummyFile),

   %% make the dummy file look loaded (profile_data demands it)
   ( prolog:'$file loaded'(DummyFile, user, 0) ->
       true
   ; otherwise ->
       assert(prolog:'$file loaded'(DummyFile, user, 0))
   ),
   %% 970509 TableName is a pred with one unit clause per line
   %%        It is what PosGoal used to be
   %%        Now PosGoal calls TableName(Line) instead.
   pos_pred_heads(Path, PosPredHead, PosTablePredHead),

   %% used by trace hook to recognize the line markers
   retractall('pos marker goal'(PosPredHead)),  % added 970506
   assert('pos marker goal'(PosPredHead)),
   %% 970509
   retractall('pos marker table goal'(PosPredHead, _)),
   assert('pos marker table goal'(PosPredHead, PosTablePredHead)),

   %% compile one clause per line of the table pred
   ( nth(ClauseNo, LinesSet, Line),
     %% assertz(user:'$ pos'(Line, Path)),
     src_pos_table_goal(Line, Path, PosTableGoal),
     compile_profiled_clause(PosTableGoal),
     %% 970516 Use this version instead if adding a terminating
     %%        default clause.
     %% Add cut so we can have a default clause.
     %% compile_profiled_clause((PosTableGoal :- !)),

     PosTableGoal = _User:PosTableHead,     
     %% for profiling to know what clause of the table is what line
     assert('line mapping'(PosTableHead, ClauseNo)),
     fail
   ; 
       %% 970516 Add the following two lines and the cut-version above
       %%        to make sure that redefinition will never cause
       %%        a trace in progress to fail due to redefined line
       %%        clauses.
       %% %% 970516 Add default clause so it will never fail.
       %% src_pos_table_goal(_OtherLine, Path, PosTableGoal),
       %% compile_profiled_clause((PosTableGoal)),
       true
   ),
   
   define_default_line_indirection(Path),
   
   %% make the pos-table goal belong to the faked file. This is because
   %% profile_data and reset_profile_data demands it (in 3PL3)
   %% MUST be done after compile_profiled_clause as it retracts the
   %% def of '$source file'
   ( LinesSet = [_|_] ->
       ( prolog:'$source file'(PosTablePredHead, user, DummyFile) ->
           %% 960901 I believe this will never happen
           true
       ; otherwise ->
           assert(prolog:'$source file'(PosTablePredHead, user, DummyFile))
       )
   ; true
   ).

%% 970503
breakable_lines(File0, Lines) :-
   %% File0 should be an absolute file name!
   %% absolute_file_name(File0, File),
   File = File0,
   %% pos_pred_heads(File, _, PosTableHead),
   src_pos_table_goal1(Line, File, PosTableHead),
   %% PosGoal =.. [File,Line],
   %% arg(PosTableHead, 1, Line),
   findall(Line, 'line mapping'(PosTableHead, _ClauseNo), Lines).


%% 970509 only exact line matches (now we have visual clues in emacs).
breakable_line(Path, Line, BreakLine) :-
   src_pos_table_goal1(Line, Path, PosTableHead),
   'line mapping'(PosTableHead, _ClauseNo), !,
   BreakLine = Line.


line_pred_dummy_file('/A nonexisting dummy file for line profiler.pl').


act_final('konsult'(_Info)).


lookup_info(Key, Val, Info) :-
   memberchk(Key-V, Info),
   Val = V.


lookup_info_swap(Key, Default, Val0, Val, Info0, Info) :-
   ( select(Key-V, Info0, Info1) ->
       Info = [Key-Val|Info1]
   ; otherwise ->
       Info = [Key-Val|Info0],
       V = Default
   ),
   Val0 = V.


compress_line_list(Ls, Cs) :- Ls = [],
   Cs = [].

compress_line_list([L|Ls], [Skip,Count|Cs]) :-
   Skip is L-1,                                 % skip from line 1
   Next is L+1,
   compress_line_list1(Ls, Next, 1, Count, Cs).

compress_line_list1([], _Next, Count0,Count, Cs) :-
   Count = Count0,
   Cs = [].

compress_line_list1([L|Ls], Next, Count0,Count, Cs) :- L == Next, !,
   Next1 is Next+1,
   Count1 is Count0+1,
   compress_line_list1(Ls, Next1, Count1,Count, Cs).

compress_line_list1([L|Ls], Next, Count0,Count, [Skip,Count1|Cs]) :-
   Count = Count0,
   Skip is (L-Next)+1,
   L1 is L+1,
   compress_line_list1(Ls, L1, 1,Count1, Cs).
%%% Copyright (c) 1995 Per Mildner, all rights reserved.
%%% email: Per.Mildner@csd.uu.se
%%% Consider adding an option to insert pos info after the last goal
%%% too, this would loose TRO but would restore source context when
%%% returning from the last call.


identical_term(T1, T2) :-
   %% Avert your eyes.
   prolog:'$bdd_equiv'(T1,T2).

%% 960901 Added Lines
%% 960809 the arg order is strange.
%% src_pos_wrap(Term:in, SrcPos:in, KConj:out, Conj:in)
src_pos_wrap(Term, File-SrcPos, KConj, Conj, Lines, LinesTail) :-
   src_pos_wrap1(SrcPos, Term, File, KConj, Conj, Lines, LinesTail).


src_pos_wrap1([], _Term, _File, KConj, Conj, Lines, Lines) :-
   KConj = Conj.

src_pos_wrap1([pos(P,T)|PTs], Term, File, KConj, Conj, Lines, LinesTail) :-
   ( identical_term(T, Term) ->
       P1 is P+1,                               % stream pos is zero-based.
       %% KConj = (source_debug:pos(P1,File),Conj)
       %% Big speed win for user code
       %% the explicit module-prefix will go away
       %% and nrev 100 goes from 780 ms to 480 ms.
       Lines = [P1|LinesTail],
       src_pos_goal(P1, File, PosGoal),
       KConj = (PosGoal,Conj)
   ; otherwise ->
       src_pos_wrap1(PTs, Term, File, KConj, Conj, Lines, LinesTail)
   ).


src_translate_body(Body, _SrcPos, KBody, Lines, LinesTail) :- var(Body), !,
   KBody = Body,
   LinesTail = Lines.

%% 960809 no need to pass double arg, SICStus will do indexing righ
%%        for clauses like foo(X) :- X = g(Y), ....
%% src_translate_body(Body, SrcPos, KBody) :-
%%   src_translate_body(Body, Body, SrcPos, KBody).

src_translate_body(Body, SrcPos, KBody, Lines, LinesTail) :- Body = (A,B), !,
   KBody = (KA,KB),
   %% Currently this has no effect as xread needs to be modified to
   %% add source-pos to ','/2 terms 
   %% 960809 Special case zero-arity tests, eg, !/0, se next clause.
   %% We assume that input looks like
   %% A,
   %% B...
   ( atom(A) ->
       %% use src pos of comma
       src_pos_wrap(Body, SrcPos, KA, KA1, Lines, Lines1)
   ; otherwise ->
       KA1 = KA,
       Lines1 = Lines
   ),
   src_translate_body(A, SrcPos, KA1, Lines1, Lines2),
   src_translate_body(B, SrcPos, KB, Lines2, LinesTail).

src_translate_body(Body, SrcPos, KBody, Lines, LinesTail) :- Body = (A->B), !,
   KBody = (KA->KB),
   %% 960809 special case zero-arity tests, eg, otherwise/0, as they
   %%        will not have any src-pos info of their own

   %%        A problem WAS (fixed 960822) that if there is a newline
   %%        immediately after the arrow (ie the usual case) then the
   %%        line-count would have increased when the reader stores
   %%        the line-count for the arrow, thus causing us to be off
   %%        by one.

   ( atom(A) ->                                 % eg otherwise
       %% use src pos of arrow
       src_pos_wrap(Body, SrcPos, KA, KA1, Lines, Lines1)
   ; otherwise ->
       KA1 = KA,
       Lines1 = Lines
   ),
   src_translate_body(A, SrcPos, KA1, Lines1, Lines2),
   src_translate_body(B, SrcPos, KB, Lines2, LinesTail).

src_translate_body((A;B), SrcPos, KBody, Lines, LinesTail) :- !,
   KBody = (KA;KB),
   %% src_pos_wrap(Body, SrcPos, KA, KA1),
   src_translate_body(A, SrcPos, KA, Lines, Lines1),
   src_translate_body(B, SrcPos, KB, Lines1, LinesTail).

src_translate_body(Goal, SrcPos, KBody, Lines, LinesTail) :- compound(Goal), !,
   %% a real goal with args (we ignore source pos info for args to call et al.)
   src_pos_wrap(Goal, SrcPos, KBody, Goal, Lines, LinesTail).

src_translate_body(Goal, _SrcPos, Goal, Lines, Lines).

%% 960902 silly
%% src_pos_goal(Line, Path, PosGoal, [Line|LinesTail], LinesTail) :-
%%    src_pos_goal(Line, Path, PosGoal).
   

%% %% temporary, used as recognizer in act.pl
%% src_pos_goal(Line, Path, user:'$ pos'(Line, Path)).

%% 970509 Note that we now also have srs_pos_table_goal/1 consider
%%
%% if changed, fix sicstus_emacs_dbg too. 960902 AND A LOT OF OTHER PLACES!
%% src_pos_goal(?Line, +Path, -PosGoal).
%% 970509 moved to act.pl
%% src_pos_goal(Line, Path, user:PosGoal) :-
%%    functor(PosGoal, Path, 1),
%%    arg(1,PosGoal,Line).
%% 9060901 new version using the built in profile support
%%
%% use source debug annotations to gather profile counts

/*
export (from load.pl and _prefix.pl):

reset_profile_data. % reset all counters

reset_profile_data(+File). % reset counters for file File

dump_profile_data.  % make emacs annotate the konsulted files.

dump_profile_data(+File).  % make emacs annotate the konsulted file.

Gone: profile_enable(-Old, +New). 'on' or 'off'

*/

:- use_module(library(lists)).


profile_enable(Old, _New) :-
   Old = on.

reset_profile_data :-
   reset_profile_data_pl5.

reset_profile_data(File) :-
   reset_profile_data_pl5(File).


pos_goals_for_profile(user:Preds) :-
   %% 970911
   findall(Name/1, ('pos marker table goal'(_PosGoalHead, PosTablePredHead),
                    functor(PosTablePredHead, Name, _)), Preds).
   %% old break scheme
   %% findall(File/1, ('pos marker goal'(PosGoalHead), functor(PosGoalHead, File, _)), Preds).


%% SICStus 3PL5
reset_profile_data_pl5 :-
   pos_goals_for_profile(Heads),
   profile_reset(Heads).

reset_profile_data_pl5(File0) :-
   absolute_file_name(File0,File),              % will suffix .pl if needed
   %% old breaks: functor(Head, File, 1),
   src_pos_table_goal1(_, File, Head),

   %% avoid warning
   ( predicate_property(user:Head, compiled) -> % 970511 was uset:, never worked!
       functor(Head,Name,_),
       profile_reset(user:(Name/1))
   ; true
   ).


%% SICStus 3PL3
reset_profile_data_pl3 :-
   line_pred_dummy_file(DummyFile),
   profile_reset(user:DummyFile).

reset_profile_data_pl3(File0) :-
   absolute_file_name(File0,File),              % will suffix .pl if needed
   Module = user,
   functor(Pred, File, 1),
   %% from prolog:profile_reset/1 in Bips/gauge.pl
   (
      prolog:predicate_counters(Pred, _, Module, _, CounterRef, Count),
      prolog:'$reset_counters'(CounterRef, Count),
      fail
   ;
      true
   ).

%% Dumps all files on one bracketed line
dump_profile_data :-
   dump_profile_data_pl5.


dump_profile_data_pl5 :-
   pos_goals_for_profile(Heads),
   profile_data(Heads, calls, clause, Data),
   dump_profile_data1(Data).


dump_profile_data_pl3 :-
   line_pred_dummy_file(DummyFile),

   profile_data(user:DummyFile, calls, clause, Data),
   dump_profile_data1(Data).
/*
   build_profile_data(Data, PerFileData),
   (
      format(user_error, '~N10L{PROFILE: (', []),
      %% BT
      member(Path-LineCounts, PerFileData),

      %% don't show un-executed files
      ( ( member(_Line-Count, LineCounts), Count > 0 ) -> true),
      dump_profile_data(Path, LineCounts),
      fail
   ;  format(user_error, ')}10R\n', [])
   ).
*/


dump_profile_data1(Data) :-
   build_profile_data(Data, PerFileData),
   (
      format(user_error, '~N10L{PROFILE: (', []),
      %% BT
      member(Path-LineCounts, PerFileData),

      %% don't show un-executed files
      ( ( member(_Line-Count, LineCounts), Count > 0 ) -> true),
      dump_profile_data(Path, LineCounts),
      fail
   ;  format(user_error, ')}10R\n', [])
   ).


build_profile_data([], []).

build_profile_data(Data, PerFileData) :-
   Data = [(user:Name/A/_)-_|_],
   build_profile_data(Data, Name/A, PerFileData).


build_profile_data(Data, FA, [Path-PathData|PerFileData]) :-
   pos_table_path(FA, Path),                    % 970511
   build_profile_data(Data, FA, PathData, PerFileData).


build_profile_data([], _FA, PathData, PerFileData) :-
   PathData = [],
   PerFileData = [].

build_profile_data(Data, FA, PathData, PerFileData) :-
   Data = [(user:FA1/ClauseNo)-Count|Data1],
   ( FA1 == FA ->
       FA = F/A,
       functor(Head, F, A),
       'line mapping'(Head, ClauseNo), !,
       arg(1, Head, LineNo),
       PathData = [LineNo-Count|PathData1],
       build_profile_data(Data1, FA, PathData1, PerFileData)
   ; otherwise ->
       PathData = [],
       %% 970511 don't duplicate the code
       %% PerFileData = [Path1-PathData1|PerFileData1],
       %% build_profile_data(Data, Path1, PathData1, PerFileData1)
       build_profile_data(Data, PerFileData)
   ).


dump_profile_data(File, LineCounts) :-
   format(user_error, '("~p" ', [File]),
   dump_profile_counts(LineCounts),
   write(user_error, ')').   


dump_profile_counts([]).

dump_profile_counts([L-C|LCs]) :-
   ( C > 0 ->
       format(user_error, '(~w ~w)', [L, C])
   ; true
   ),
   dump_profile_counts(LCs).


profile_file_mappings(File, Files) :- var(File), !,
   Files = [File-File].

profile_file_mappings(File0, Files) :- atom(File), !,
   absolute_file_name(File0,File),
   Files = [File-File].

profile_file_mappings(FileMappings0, FileMappings) :-
   setof(X-Y, (member(X0-Y0, FileMappings0),
               absolute_file_name(X0,X), !,
               absolute_file_name(Y0,Y), !
              ), FileMappings).
   


/*
dump_profile_data :-
   format('~N;;; Start of file dumps~n', []),
   profile_data(File, LineCounts),              % backtrack over files
   dump_profile_data(File, LineCounts),
   fail.

dump_profile_data :-
   format('~N;;; End of file dumps~n', []).


dump_profile_data(File, LineCounts) :-
   format('~N(~n;;; Profile data~n   "~p"~n', [File]),
   dump_profile_counts(LineCounts),
   format('~N;;; End of profile data for ~p~n)~n', [File]).


dump_profile_counts([]).

dump_profile_counts([L-C|LCs]) :-
   format('~n   (~w ~w)', [L, C]),
   dump_profile_counts(LCs).
*/

/*
source_debug_profile(on) :- !,
   retractall(source_debug:pos(_,_)),
   assert((source_debug:pos(Line, File) :- pos_profile(Line, File))).

source_debug_profile(off) :-
   retractall(source_debug:pos(_,_)),
   assert((source_debug:pos(_Line, _Pos))).
*/

/* 960901 now gone, unfortunately
profile_enable(Old, New) :-
   Flag = profile_flag,
   %% Flag = prolog:sicstus_debug_profile_flag,
   ( call(Flag) ->
     Old = on
   ; Old = off
   ),
   ( New == Old ->
       true
   ; otherwise ->
       ( New == on ->
           retractall(Flag),
           assert(Flag)
       ; New == off ->
           retractall(Flag)
       ),
       ( New = on ->
           retractall(user:'$ pos'(_,_)),                % NOT abolish! (preserve spy point)
           assert((user:'$ pos'(Line,File) :- pos_profile(Line, File)))
       ; New = off ->
           retractall(user:'$ pos'(_,_)),                % NOT abolish! (preserve spy point)
           assert((user:'$ pos'(_Line,_File)))
       )
   ).
*/

%% 960901 rewriting to use built-in profiler and one clause per line.
%% %% By changing the def dynamically we can avoid seing the call to
%% %% profile_flag when tracing un-profiled code.
%% user:'$ pos'(_Line, _File).

/*
%% The module prefix necessary for pos/2 makes nrev go from 410ms to 760 ms!
user:'$ pos'(Line, File) :- profile_flag, !,          % 760 ms
   pos_profile(Line, File).

user:'$ pos'(_Line, _File).
*/



%% %% The bad thing with this is that the call to profile_flag shows up when tracing
%% pos(Line, File) :- profile_flag, !,          % 760 ms
%%    pos_profile(Line, File).
%%
%% pos(_Line, _File).

%% Various slower alternatives:
%% pos(Line, File) :- prolog:clause_internal(sicstus_debug_profile_flag, _Ptr), !, % 990 ms
%%   pos_profile(Line, File).
%% pos(Line, File) :- clause(profile_flag, _), !,  % 1240 ms
%%    pos_profile(Line, File).
%% pos(_Line, _File) :- no_profile_flag, !. % 860 ms
%% pos(_Line, _File) :- prolog:clause_internal(sicstus_debug_no_profile_flag, _Ptr), !.% 990 ms
%% pos(Line, File) :-
%%    pos_profile(Line, File).

/* 960901
reset_profile_data :-
   reset_profile_data(_).

reset_profile_data(File) :-
   retractall(profile_pos(_, File, _)).
*/

/*
pos_profile(Line, File) :-
   ( clause(profile_pos(Line, File, Count), true, Ref) ->
       erase(Ref),
       Count1 is Count+1
   ; otherwise ->
       Count1 = 1
   ),
   assert(profile_pos(Line, File, Count1)).
*/


%% maybe somewhat faster but not by much (10%)
%%%  pos_profile1(Line, File) :- clause(profile_pos(Line, File, Count), true), !,
%%%     ( 0 is Count mod 100 ->
%%%         abolish(profile_pos, 3)
%%%     ; true
%%%     ),
%%%     Count1 is Count+1,
%%%     asserta(profile_pos(Line, File, Count1)).
%%%  
%%%  pos_profile1(Line, File) :- asserta(profile_pos(Line, File, 1)).


/* 960901
profile_data(File, LineCounts) :-
   setof(Line-Count, profile_pos(Line, File, Count), LineCounts).


%% This version dumps all files on one bracketed line
dump_profile_data :-                            % 960805
   dump_profile_data(_AnyFile).


dump_profile_data(File0) :-
   profile_file_mappings(File0, Files),
   format(user_error, '~N10L{PROFILE: (', []),
   member(RealFile-OtherName, Files),
   profile_data(RealFile, LineCounts),              % backtrack over files
   dump_profile_data(OtherName, LineCounts),
   fail.

dump_profile_data(_File) :-
   format(user_error, ')}10R', []),
   flush_output(user_error).
*/

%% :- dynamic profile_pos/3.
%% :- dynamic profile_flag/0.
%% :- dynamic no_profile_flag/0.
%% :- dynamic prolog:sicstus_debug_profile_flag/0.

%% 960901 now replaced by user:Path(LineNumber)
%% :- dynamic user:'$ pos'/2.


%% no_profile_flag.
%% 960901 compile_clause/1
/* Some context is needed for best results, it could probably be
   ignored but is needed for gauge-style profiling to work on the
   compiled clauses.

   %% Do this once
   DummyFile = 'a dummy file we pretend the compiled clauses comes from',
   ( prolog:'$file loaded'(DummyFile, user, 0) ->
       true
   ; otherwise ->
       assert(prolog:'$file loaded'(DummyFile, user, 0))
   ),
   GoalHead = foo(_,_,_),
   %% compile_comact_clause/1 or compile_profiled_clause/1
   compile_compact_clause((foo([],B,C) :- ...)),
   compile_compact_clause((foo([A|As],B,C) :- ...)),
   compile_compact_clause((foo(A,B,C) :- ...)),
   
   %% must be done AFTER all clauses for a incore-compiled predicate
   %% have been compiled with compile_clause/1
   assert(prolog:'$source file'(PosGoalHead, user, DummyFile)

*/





compile_profiled_clause(Clause) :-
   prolog_flag(compiling, Old),
   on_exception(E,
                (prolog_flag(compiling, _, profiledcode),
                 compile_clause(Clause)),
                (prolog_flag(compiling, _, Old),
                 raise_exception(E))),
   prolog_flag(compiling, _, Old).


compile_compact_clause(Clause) :-
   prolog_flag(compiling, Old),
   on_exception(E,
                (prolog_flag(compiling, _, compactcode),
                 compile_clause(Clause)),
                (prolog_flag(compiling, _, Old),
                 raise_exception(E))),
   prolog_flag(compiling, _, Old).


%% compile_clause(:Clause):
%% adds a single Clause to the database, without term expansion.
%% If the predicate already exists, Clause is added to the current def.
%% Otherwise, a new predicate is created according to the value of the
%% 'compiling' Prolog flag (compiling must not be fastcode!)
%% Native code predicates cannot be modified.
%% Predicates in the prolog module cannot be modified.
compile_clause(M:C) :- !,
   compile_clause1(C,M).
compile_clause(C) :-
   prolog_flag(typein_module, M),
   compile_clause1(C,M).


compile_clause1(M:C, _M) :- !,
   compile_clause1(C, M).

compile_clause1((Head0 :- Body), Module0) :- !,
        prolog:get_module(Head0, Module0, Head, Module),
        callable(Head),
        prolog:exp_vars(Head, HV, Module, incore),
        prolog:wellformed_body(Body, +, Body1, HV, Module0, Module, incore),
        shell_pred(Head, Module, Mode),
        (   Mode==ignore
        ->  functor(Head, F, A),
            print_message(warning, not_loaded(F/A,compiled))
        ;   prolog:shell_step(Mode, Head, Body1, Module, user)
        ).
compile_clause1(Head, M) :-
        compile_clause1((Head :- true), M).


shell_pred(Head, Module, Mode) :-
        prolog:'$predicate_property'(Head, Type, Props, Module, HomeModule), !,
        (   HomeModule==prolog ->  Mode = ignore
        ;   Props/\8 =:= 8 ->  Mode = multifile

%        ;   prolog:shell_enter_mode1(Type, Mode), Mode\==fastcode -> true 
% if using PL6, use this line instead of the one above:
        ;   my_shell_enter_mode1(Type, Mode), Mode\==fastcode -> true 

        ;   Mode = ignore
        ).
shell_pred(Head, Module, Mode) :-
        prolog:abolish_defined_in(Head, Module),
        prolog_flag(compiling, Compiling),
        (   Compiling==fastcode -> Mode = ignore
        ;   Mode = Compiling,
            functor(Head, F, A),
            prolog:'$define_predicate'((Module:F)/A, Compiling)
        ).

% if using PL6, uncomment these definitions
   my_shell_enter_mode1(0, compactcode).		% COMPACTCODE
   my_shell_enter_mode1(1, compactcode).		% COMPACTCODE_INDEXED
   my_shell_enter_mode1(2, profiledcode).		% PROFILEDCODE
   my_shell_enter_mode1(3, profiledcode).		% PROFILEDCODE_INDEXED
   my_shell_enter_mode1(4, fastcode).			% FASTCODE
   my_shell_enter_mode1(5, fastcode).			% FASTCODE_INDEXED
   my_shell_enter_mode1(8, debugcode).			% INTERPRETED

